package com.ruoyi.sim.service.impl;

import com.ruoyi.sim.domain.SimMsg;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import static com.ruoyi.sim.service.impl.CommConst.*;

@Service
public class CommBuildService {


    @Autowired
    private SnowflakeIdService idService;

    /**
     * 设备类型读取
     *
     * @param simNum sim.sim_num
     */
    public SimMsg buildSendMsgReadSimType(final String simNum) {
        return buildSendMsg(simNum, CMD_READ_TYPE, CMD_ID_GET_SN);
    }

    /**
     * 故障下发
     *
     * @param simNum          sim.sim_num
     * @param bindHardwareMsg fault.bind_hardware_msg
     */
    public SimMsg buildSendMsgWriteFault(final String simNum, final String bindHardwareMsg) {
        return buildSendMsg(simNum, CommConst.CMD_SET_FAULT, bindHardwareMsg);
    }

    /**
     * 状态读取
     *
     * @param simNum          sim.sim_num
     * @param bindHardwareMsg fault.bind_hardware_msg
     */
    public SimMsg buildSendMsgReadFaultResistance(final String simNum, final String bindHardwareMsg) {
        return buildSendMsg(simNum, CMD_READ_FAULT_RESISTANCE, bindHardwareMsg);
    }

    /**
     * 故障清清除
     *
     * @param simNum          sim.sim_num
     * @param bindHardwareMsg fault.bind_hardware_msg
     */
    public SimMsg buildSendMsgClearFault(final String simNum, final String bindHardwareMsg) {
        return buildSendMsg(simNum, CMD_CLEAR_FAULT, bindHardwareMsg);
    }

    public SimMsg buildSendMsg(final String simNum, final String cmd, final String cmdId) {
        return buildSendMsg(simNum, cmd, cmdId, CMD_DATA_PLACE_HOLDER);
    }

    public SimMsg buildSendMsg(final String simNum, final String cmd, final String cmdId, final String data) {
        SimMsg smS = new SimMsg();
        smS.setSimId(idService.nextId());
        if (StringUtils.isEmpty(simNum) || StringUtils.isEmpty(cmd) || StringUtils.isEmpty(cmdId) || StringUtils.isEmpty(data)) {
            throw new IllegalArgumentException("buildSendMsg isEmpty");
        }
        if (simNum.length() != LENGTH_2) {
            throw new IllegalArgumentException("buildSendMsg length error");
        }
        if (cmd.length() != LENGTH_2) {
            throw new IllegalArgumentException("buildSendMsg length error");
        }
        if (cmdId.length() != LENGTH_2) {
            throw new IllegalArgumentException("buildSendMsg length error");
        }
        if (data.length() != LENGTH_8) {
            throw new IllegalArgumentException("buildSendMsg length error");
        }
        StringBuffer m = new StringBuffer();
        m.append(PREFIX);
        m.append(simNum);
        m.append(ORN_SEND);
        m.append(cmd);
        m.append(cmdId);
        m.append(data);
        m.append(SUFFIX);
        final String mFinal = m.toString();
        if (mFinal.length() != LENGTH_20) {
            throw new IllegalArgumentException("buildSendMsg length error");
        }
        smS.setSendMsg(mFinal);
        return smS;
    }

    /**
     * check receiveMsg
     *
     * @param receiveMsg
     * @return
     */
    public boolean checkReceiveMsg(String receiveMsg) {
        if (StringUtils.isEmpty(receiveMsg)) {
            return false;
        }
        if (receiveMsg.length() != LENGTH_20) {
            return false;
        }
        final String orn = StringUtils.substring(receiveMsg, 4, 6);
        if (!ORN_RECEIVE.equals(orn)) {
            return false;
        }
        if (!StringUtils.startsWith(receiveMsg, PREFIX)) {
            return false;
        }
        if (!StringUtils.endsWith(receiveMsg, SUFFIX)) {
            return false;
        }
        return true;
    }
}
